@@ -75,6 +75,8 @@ struct bgp_af_desc {
 
 struct bgp_config {
   struct proto_config c;
+  /* Beginning here, all options must be comparable by memcmp().
+   * Do not add anything above unless you want this comment to become futile. */
   u32 local_as, remote_as;
   ip_addr local_ip;			/* Source address to use */
   ip_addr remote_ip;
@@ -85,6 +87,7 @@ struct bgp_config {
   int multihop;				/* Number of hops if multihop */
   int strict_bind;			/* Bind listening socket to local address */
   int free_bind;			/* Bind listening socket with SKF_FREEBIND */
+  int onlink;				/* Enable direct connection to a host not configured on any iface */
   int ttl_security;			/* Enable TTL security [RFC 5082] */
   int compare_path_lengths;		/* Use path lengths when selecting best route */
   int med_metric;			/* Compare MULTI_EXIT_DISC even between routes from differen ASes */
@@ -102,6 +105,7 @@ struct bgp_config {
   u32 rr_cluster_id;			/* Route reflector cluster ID, if different from local ID */
   int rr_client;			/* Whether neighbor is RR client of me */
   int rs_client;			/* Whether neighbor is RS client of me */
+  int ipv4;				/* Use IPv4 connection, i.e. remote_ip is IPv4 */
   u32 confederation;			/* Confederation ID, or zero if confeds not active */
   int confederation_member;		/* Whether neighbor AS is member of our confederation */
   int passive;				/* Do not initiate outgoing connection */
@@ -143,6 +147,7 @@ struct bgp_config {
   const char *password;			/* Password used for MD5 authentication */
   struct ao_config *ao_keys;		/* Keys for TCP-AO authentication */
   net_addr *remote_range;		/* Allowed neighbor range for dynamic BGP */
+  struct iface_patt *ipatt;		/* Interface patterns for dynamic strict bind */
   const char *dynamic_name;		/* Name pattern for dynamic BGP */
   int dynamic_name_digits;		/* Minimum number of digits for dynamic names */
   int check_link;			/* Use iface link state for liveness detection */
@@ -307,12 +312,37 @@ struct bgp_ao_state {
   struct bgp_ao_key *best_key;
 };
 
-struct bgp_socket {
-  node n;				/* Node in global bgp_sockets */
+struct bgp_socket_params {
+  ip_addr addr;				/* Local address to bind to */
+  struct iface *iface;			/* Local interface to bind to */
+  struct iface *vrf;			/* VRF to bind to */
+  uint port;				/* Local port to bind to (mandatory) */
+  uint flags;				/* Additional SKF_* flags */
+};
+
+#define BGP_SOCKET_PUB \
+  node n;				/* Node in global bgp_listen -> sockets */ \
+  struct birdloop *loop;		/* Socket's accepting loop */ \
+  struct bgp_socket_params params;	/* Socket matching parameters */ \
+
+struct bgp_socket_private {
+  BGP_SOCKET_PUB;
+  struct bgp_socket_private **locked_at;
   list requests;			/* Listen requests */
   sock *sk;				/* Real listening socket */
 };
 
+typedef union bgp_socket {
+  struct { BGP_SOCKET_PUB; };
+  struct bgp_socket_private priv;
+} bgp_socket;
+
+BLO_UNLOCK_CLEANUP(bgp_socket);
+
+#define BGP_SOCKET_LOCKED(_pub, _priv)	BLO_LOCKED(_pub, _priv, bgp_socket)
+#define BGP_SOCKET_LOCK(_pub, _priv)	BLO_LOCK(_pub, _priv, bgp_socket)
+#define BGP_SOCKET_UNLOCK(_pub, _priv)	( BLO_UNLOCK_CLEANUP_NAME(bgp_socket)(_priv), _priv = NULL )
+
 struct bgp_stats {
   uint rx_messages, tx_messages;
   uint rx_updates, tx_updates;
@@ -366,14 +396,24 @@ struct bgp_session_close_ad {
   byte data[0];
 };
 
+struct bgp_incoming_socket {
+  node n;		/* Node in bgp_listen_request -> incoming_sockets */
+  sock *sk;		/* The actual socket */
+};
+
 struct bgp_listen_request {
-  node n;				/* Node in bgp_socket / pending list */
-  struct bgp_socket *sock;		/* Assigned socket */
-  ip_addr addr;
-  struct iface *iface;
-  struct iface *vrf;
-  uint port;
-  uint flags;
+  node pn;				/* Node in bgp_proto listen list */
+  node sn;				/* Node in bgp_socket requests list */
+  bgp_socket *sock;			/* Assigned socket */
+  struct bgp_socket_params params;	/* Listening socket parameters */
+  ip_addr local_ip;			/* Local IP address to match */
+  struct iface *iface;			/* Local interface to match */
+  struct iface_patt *ipatt;		/* Interface pattern for dynamic strict bind */
+  ip_addr remote_ip;			/* Remote IP address to match */
+  const net_addr *remote_range;		/* Remote IP range to match */
+  list incoming_sockets;		/* Accepted sockets matched to this request */
+  callback incoming_connection;		/* Callback for incoming connection */
+  struct bgp_proto *p;			/* Requesting protocol */
 };
 
 struct bgp_proto {
@@ -407,7 +447,7 @@ struct bgp_proto {
   struct bgp_conn incoming_conn;	/* Incoming connection we have neither accepted nor rejected yet */
   struct object_lock *lock;		/* Lock for neighbor connection */
   struct neighbor *neigh;		/* Neighbor entry corresponding to remote ip, NULL if multihop */
-  struct bgp_listen_request listen;	/* Shared listening socket */
+  list listen;				/* Requests for shared listening sockets */
   struct bfd_request_ref *bfd_req;	/* BFD request, if BFD is used */
   callback bfd_notify;			/* BFD notification callback */
   struct birdsock *postponed_sk;	/* Postponed incoming socket for dynamic BGP */
