@@ -24,6 +24,7 @@ const char *ospf_inm_names[] = {
 
 static int can_do_adj(struct ospf_neighbor *n);
 static void inactivity_timer_hook(timer * timer);
+static void ospf_neigh_bfd_hook(callback *cb);
 static void dbdes_timer_hook(timer *t);
 static void lsrq_timer_hook(timer *t);
 static void lsrt_timer_hook(timer *t);
@@ -80,7 +81,7 @@ struct ospf_neighbor *
 ospf_neighbor_new(struct ospf_iface *ifa)
 {
   struct ospf_proto *p = ifa->oa->po;
-  struct pool *pool = rp_new(p->p.pool, "OSPF Neighbor");
+  struct pool *pool = rp_new(p->p.pool, proto_domain(&p->p), "OSPF Neighbor");
   struct ospf_neighbor *n = mb_allocz(pool, sizeof(struct ospf_neighbor));
 
   n->pool = pool;
@@ -100,6 +101,8 @@ ospf_neighbor_new(struct ospf_iface *ifa)
   n->lsrt_timer = tm_new_init(pool, lsrt_timer_hook, n, ifa->rxmtint S, 0);
   n->ackd_timer = tm_new_init(pool, ackd_timer_hook, n, ifa->rxmtint S / 2, 0);
 
+  callback_init(&n->bfd_notify, ospf_neigh_bfd_hook, p->p.loop);
+
   return (n);
 }
 
@@ -120,7 +123,7 @@ ospf_neigh_down(struct ospf_neighbor *n)
   s_get(&(n->dbsi));
   release_lsrtl(p, n);
   rem_node(NODE n);
-  rfree(n->pool);
+  rp_free(n->pool);
 
   OSPF_TRACE(D_EVENTS, "Neighbor %R on %s removed", rid, ifa->ifname);
 }
@@ -186,8 +189,8 @@ ospf_neigh_chstate(struct ospf_neighbor *n, u8 state)
     n->myimms = DBDES_IMMS;
     n->got_my_rt_lsa = 0;
 
-    tm_start(n->dbdes_timer, 0);
-    tm_start(n->ackd_timer, ifa->rxmtint S / 2);
+    tm_start_in(n->dbdes_timer, 0, p->p.loop);
+    tm_start_in(n->ackd_timer, ifa->rxmtint S / 2, p->p.loop);
   }
 
   if (state > NEIGHBOR_EXSTART)
@@ -232,7 +235,7 @@ ospf_neigh_sm(struct ospf_neighbor *n, int event)
       ospf_neigh_chstate(n, NEIGHBOR_INIT);
 
     /* Restart inactivity timer */
-    tm_start(n->inactim, n->ifa->deadint S);
+    tm_start_in(n->inactim, n->ifa->deadint S, p->p.loop);
     break;
 
   case INM_2WAYREC:
@@ -382,7 +385,7 @@ ospf_neigh_start_graceful_restart(struct ospf_neighbor *n, uint gr_time)
   p->gr_count++;
 
   n->gr_timer = tm_new_init(n->pool, graceful_restart_timeout, n, 0, 0);
-  tm_start(n->gr_timer, gr_time S);
+  tm_start_in(n->gr_timer, gr_time S, p->p.loop);
 }
 
 static void
@@ -493,7 +496,7 @@ ospf_neigh_notify_grace_lsa(struct ospf_neighbor *n, struct top_hash_entry *en)
     /* Exception for updating grace period */
     if (n->gr_active)
     {
-      tm_start(n->gr_timer, (period S) - (en->lsa.age S));
+      tm_start_in(n->gr_timer, (period S) - (en->lsa.age S), p->p.loop);
       return;
     }
 
@@ -756,9 +759,12 @@ inactivity_timer_hook(timer * timer)
 }
 
 static void
-ospf_neigh_bfd_hook(struct bfd_request *req)
+ospf_neigh_bfd_hook(callback *cb)
 {
-  struct ospf_neighbor *n = req->data;
+  SKIP_BACK_DECLARE(struct ospf_neighbor, n, bfd_notify, cb);
+  struct bfd_request *req = n->bfd_req->req;
+  bfd_request_update_state(req);
+
   struct ospf_proto *p = n->ifa->oa->po;
 
   if (req->down)
@@ -777,7 +783,7 @@ ospf_neigh_update_bfd(struct ospf_neighbor *n, int use_bfd)
   if (use_bfd && !n->bfd_req)
     n->bfd_req = bfd_request_session(n->pool, n->ip, n->ifa->addr->ip,
 				     n->ifa->iface, p->p.vrf,
-				     ospf_neigh_bfd_hook, n, NULL);
+				     &n->bfd_notify, NULL);
 
   if (!use_bfd && n->bfd_req)
   {
