@@ -538,7 +538,7 @@ radv_sk_open(struct radv_iface *ifa)
   sk->data = ifa;
   sk->flags = SKF_LADDR_RX;
 
-  if (sk_open(sk) < 0)
+  if (sk_open(sk, ifa->ra->p.loop) < 0)
     goto err;
 
   /* We want listen just to ICMPv6 messages of type RS and RA */
@@ -556,7 +556,7 @@ radv_sk_open(struct radv_iface *ifa)
 
  err:
   sk_log_error(sk, ifa->ra->p.name);
-  rfree(sk);
+  sk_close(sk);
   return 0;
 }
 
