@@ -436,7 +436,7 @@ static inline void
 rip_kick_rxmt_timer(struct rip_iface *ifa)
 {
   if (! tm_active(ifa->rxmt_timer))
-    tm_start(ifa->rxmt_timer, ifa->cf->rxmt_time);
+    tm_start_in(ifa->rxmt_timer, ifa->cf->rxmt_time, ifa->rip->p.loop);
 }
 
 
@@ -989,6 +989,8 @@ drop:
 int
 rip_open_socket(struct rip_iface *ifa)
 {
+  ASSERT_DIE(birdloop_inside(ifa->rip->p.loop));
+
   struct rip_proto *p = ifa->rip;
 
   sock *sk = sk_new(p->p.pool);
@@ -1012,7 +1014,7 @@ rip_open_socket(struct rip_iface *ifa)
 
   /* sk->rbsize and sk->tbsize are handled in rip_iface_update_buffers() */
 
-  if (sk_open(sk) < 0)
+  if (sk_open(sk, p->p.loop) < 0)
     goto err;
 
   if (ifa->cf->mode == RIP_IM_MULTICAST)
@@ -1040,6 +1042,6 @@ rip_open_socket(struct rip_iface *ifa)
 
 err:
   sk_log_error(sk, p->p.name);
-  rfree(sk);
+  sk_close(sk);
   return 0;
 }
