@@ -109,6 +109,7 @@ static void rpki_schedule_next_expire_check(struct rpki_cache *cache);
 static void rpki_stop_refresh_timer_event(struct rpki_cache *cache);
 static void rpki_stop_retry_timer_event(struct rpki_cache *cache);
 static void rpki_stop_expire_timer_event(struct rpki_cache *cache);
+static void rpki_stop_all_timers(struct rpki_cache *cache);
 
 
 /*
@@ -120,24 +121,20 @@ rpki_table_add_roa(struct rpki_cache *cache, struct channel *channel, const net_
 {
   struct rpki_proto *p = cache->p;
 
-  rta a0 = {
-    .pref = channel->preference,
-    .source = RTS_RPKI,
-    .scope = SCOPE_UNIVERSE,
-    .dest = RTD_NONE,
-  };
+  ea_list *ea = NULL;
+  ea_set_attr_u32(&ea, &ea_gen_preference, 0, channel->preference);
+  ea_set_attr_u32(&ea, &ea_gen_source, 0, RTS_RPKI);
 
-  rta *a = rta_lookup(&a0);
-  rte *e = rte_get_temp(a, p->p.main_source);
+  rte e0 = { .attrs = ea, .src = p->p.main_source, };
 
-  rte_update2(channel, &pfxr->n, e, e->src);
+  rte_update(channel, &pfxr->n, &e0, p->p.main_source);
 }
 
 void
 rpki_table_remove_roa(struct rpki_cache *cache, struct channel *channel, const net_addr_union *pfxr)
 {
   struct rpki_proto *p = cache->p;
-  rte_update2(channel, &pfxr->n, NULL, p->p.main_source);
+  rte_update(channel, &pfxr->n, NULL, p->p.main_source);
 }
 
 void
@@ -147,20 +144,16 @@ rpki_table_add_aspa(struct rpki_cache *cache, struct channel *channel,
   struct rpki_proto *p = cache->p;
 
   net_addr_union n = { .aspa = NET_ADDR_ASPA(customer) };
-  rta a0 = {
-    .pref = channel->preference,
-    .source = RTS_RPKI,
-    .scope = SCOPE_UNIVERSE,
-    .dest = RTD_NONE,
-  };
 
-  ea_set_attr_data(&a0.eattrs, tmp_linpool, EA_ASPA_PROVIDERS, 0,
-		   EAF_TYPE_INT_SET, providers, providers_length);
+  ea_list *ea = NULL;
+  ea_set_attr_u32(&ea, &ea_gen_preference, 0, channel->preference);
+  ea_set_attr_u32(&ea, &ea_gen_source, 0, RTS_RPKI);
 
-  rta *a = rta_lookup(&a0);
-  rte *e = rte_get_temp(a, p->p.main_source);
+  ea_set_attr_data(&ea, &ea_gen_aspa_providers, 0, providers, providers_length);
 
-  rte_update2(channel, &n.n, e, e->src);
+  rte e0 = { .attrs = ea, .src = p->p.main_source, };
+
+  rte_update(channel, &n.n, &e0, p->p.main_source);
 }
 
 void
@@ -168,9 +161,37 @@ rpki_table_remove_aspa(struct rpki_cache *cache, struct channel *channel, u32 cu
 {
   struct rpki_proto *p = cache->p;
   net_addr_union n = { .aspa = NET_ADDR_ASPA(customer) };
-  rte_update2(channel, &n.n, NULL, p->p.main_source);
+  rte_update(channel, &n.n, NULL, p->p.main_source);
 }
 
+void
+rpki_start_refresh(struct rpki_proto *p)
+{
+  if (p->roa4_channel)
+    rt_refresh_begin(&p->roa4_channel->in_req);
+  if (p->roa6_channel)
+    rt_refresh_begin(&p->roa6_channel->in_req);
+  if (p->aspa_channel)
+    rt_refresh_begin(&p->aspa_channel->in_req);
+
+  p->refresh_channels = 1;
+}
+
+void
+rpki_stop_refresh(struct rpki_proto *p)
+{
+  if (!p->refresh_channels)
+    return;
+
+  p->refresh_channels = 0;
+
+  if (p->roa4_channel)
+    rt_refresh_end(&p->roa4_channel->in_req);
+  if (p->roa6_channel)
+    rt_refresh_end(&p->roa6_channel->in_req);
+  if (p->aspa_channel)
+    rt_refresh_end(&p->aspa_channel->in_req);
+}
 
 /*
  *	RPKI Protocol Logic
@@ -227,13 +248,15 @@ rpki_force_restart_proto(struct rpki_proto *p)
 {
   if (p->cache)
   {
+    rpki_tr_close(p->cache->tr_sock);
+    rpki_stop_all_timers(p->cache);
     CACHE_DBG(p->cache, "Connection object destroying");
   }
 
   /* Sign as freed */
   p->cache = NULL;
 
-  proto_notify_state(&p->p, PS_DOWN);
+  proto_notify_state(&p->p, PS_FLUSH);
 }
 
 /**
@@ -352,7 +375,7 @@ rpki_schedule_next_refresh(struct rpki_cache *cache)
   btime t = cache->refresh_interval S;
 
   CACHE_DBG(cache, "after %t s", t);
-  tm_start(cache->refresh_timer, t);
+  tm_start_in(cache->refresh_timer, t, cache->p->p.loop);
 }
 
 static void
@@ -361,7 +384,7 @@ rpki_schedule_next_retry(struct rpki_cache *cache)
   btime t = cache->retry_interval S;
 
   CACHE_DBG(cache, "after %t s", t);
-  tm_start(cache->retry_timer, t);
+  tm_start_in(cache->retry_timer, t, cache->p->p.loop);
 }
 
 static void
@@ -372,7 +395,7 @@ rpki_schedule_next_expire_check(struct rpki_cache *cache)
   t = MAX(t, 1 S);
 
   CACHE_DBG(cache, "after %t s", t);
-  tm_start(cache->expire_timer, t);
+  tm_start_in(cache->expire_timer, t, cache->p->p.loop);
 }
 
 static void
@@ -389,13 +412,21 @@ rpki_stop_retry_timer_event(struct rpki_cache *cache)
   tm_stop(cache->retry_timer);
 }
 
-static void UNUSED
+static void
 rpki_stop_expire_timer_event(struct rpki_cache *cache)
 {
   CACHE_DBG(cache, "Stop");
   tm_stop(cache->expire_timer);
 }
 
+static void
+rpki_stop_all_timers(struct rpki_cache *cache)
+{
+  rpki_stop_refresh_timer_event(cache);
+  rpki_stop_retry_timer_event(cache);
+  rpki_stop_expire_timer_event(cache);
+}
+
 static int
 rpki_sync_is_stuck(struct rpki_cache *cache)
 {
@@ -417,6 +448,9 @@ rpki_refresh_hook(timer *tm)
 {
   struct rpki_cache *cache = tm->data;
 
+  if (cache->p->cache != cache)
+    return;
+
   CACHE_DBG(cache, "%s", rpki_cache_state_to_str(cache->state));
 
   switch (cache->state)
@@ -463,6 +497,9 @@ rpki_retry_hook(timer *tm)
 {
   struct rpki_cache *cache = tm->data;
 
+  if (cache->p->cache != cache)
+    return;
+
   CACHE_DBG(cache, "%s", rpki_cache_state_to_str(cache->state));
 
   switch (cache->state)
@@ -513,6 +550,9 @@ rpki_expire_hook(timer *tm)
 {
   struct rpki_cache *cache = tm->data;
 
+  if (cache->p->cache != cache)
+    return;
+
   if (!cache->last_update)
     return;
 
@@ -590,7 +630,7 @@ rpki_check_expire_interval(uint seconds)
 static struct rpki_cache *
 rpki_init_cache(struct rpki_proto *p, struct rpki_config *cf)
 {
-  pool *pool = rp_new(p->p.pool, cf->hostname);
+  pool *pool = rp_new(p->p.pool, proto_domain(&p->p), cf->hostname);
 
   struct rpki_cache *cache = mb_allocz(pool, sizeof(struct rpki_cache));
 
@@ -656,6 +696,7 @@ rpki_close_connection(struct rpki_cache *cache)
 {
   CACHE_TRACE(D_EVENTS, cache, "Closing a connection");
   rpki_tr_close(cache->tr_sock);
+  rpki_stop_refresh(cache->p);
   proto_notify_state(&cache->p->p, PS_START);
 }
 
@@ -667,7 +708,7 @@ rpki_shutdown(struct proto *P)
   rpki_force_restart_proto(p);
 
   /* Protocol memory pool will be automatically freed */
-  return PS_DOWN;
+  return PS_FLUSH;
 }
 
 
@@ -867,7 +908,7 @@ rpki_get_status(struct proto *P, byte *buf)
 {
   struct rpki_proto *p = (struct rpki_proto *) P;
 
-  if (P->proto_state == PS_DOWN)
+  if ((P->proto_state == PS_DOWN_XX) || (P->proto_state == PS_FLUSH))
   {
     *buf = 0;
     return;
@@ -895,7 +936,7 @@ rpki_show_proto_info(struct proto *P)
   struct rpki_config *cf = (void *) p->p.cf;
   struct rpki_cache *cache = p->cache;
 
-  if (P->proto_state == PS_DOWN)
+  if ((P->proto_state == PS_DOWN_XX) || (P->proto_state == PS_FLUSH))
     return;
 
   if (cache)
@@ -1031,10 +1072,10 @@ rpki_copy_config(struct proto_config *dest UNUSED, struct proto_config *src UNUS
 struct protocol proto_rpki = {
   .name = 		"RPKI",
   .template = 		"rpki%d",
-  .class =		PROTOCOL_RPKI,
   .preference = 	DEF_PREF_RPKI,
   .proto_size = 	sizeof(struct rpki_proto),
   .config_size =	sizeof(struct rpki_config),
+  .startup =		PROTOCOL_STARTUP_GENERATOR,
   .init = 		rpki_init,
   .start = 		rpki_start,
   .postconfig = 	rpki_postconfig,
