@@ -28,6 +28,7 @@
 #include "lib/resource.h"
 #include "lib/socket.h"
 #include "lib/event.h"
+#include "lib/locking.h"
 #include "lib/timer.h"
 #include "lib/tlists.h"
 #include "lib/string.h"
@@ -67,7 +68,7 @@ async_dump_run(struct dump_request *dreq)
   // XXXX tm_dump_all();
   if_dump_all(dreq);
   neigh_dump_all(dreq);
-  rta_dump_all(dreq);
+  ea_dump_all(dreq);
   rt_dump_all(dreq);
   protos_dump_all(dreq);
 
@@ -190,6 +191,8 @@ read_iproute_table(struct config *conf, char *file, char *prefix, uint max)
 #endif // PATH_IPROUTE_DIR
 
 
+#define CONFIG_COMMIT(cf, ...)	({ CONFIG_REF_LOCAL(cr, cf); config_commit(&cr, __VA_ARGS__); })
+
 static char *config_name = PATH_CONFIG_FILE;
 
 static int
@@ -223,12 +226,14 @@ sysdep_preconfig(struct config *c)
 
 static void cli_commit(struct config *new, struct config *old);
 
-int
+void
 sysdep_commit(struct config *new, struct config *old)
 {
-  log_switch(0, &new->logfiles, new->syslog_name);
+  if (!new->shutdown)
+    log_switch(0, &new->logfiles, new->syslog_name);
+
   cli_commit(new, old);
-  return 0;
+  bird_thread_commit(new, old);
 }
 
 static int
@@ -280,7 +285,7 @@ async_config(void)
       config_free(conf);
     }
   else
-    config_commit(conf, RECONFIG_HARD, 0);
+    CONFIG_COMMIT(conf, RECONFIG_HARD, 0);
 }
 
 static struct config *
@@ -361,7 +366,7 @@ cmd_reconfig(const char *name, int type, uint timeout)
   if (!conf)
     return;
 
-  int r = config_commit(conf, type, timeout);
+  int r = CONFIG_COMMIT(conf, type, timeout);
 
   if ((r >= 0) && (timeout > 0))
     {
@@ -446,7 +451,8 @@ static TLIST_LIST(cli_listener) cli_listeners;
 static void
 cli_write(cli *c)
 {
-  sock *s = c->priv;
+  sock *s = c->sock;
+  ASSERT_DIE(c->sock);
 
   while (c->tx_pos)
     {
@@ -470,7 +476,9 @@ cli_write(cli *c)
 void
 cli_write_trigger(cli *c)
 {
-  sock *s = c->priv;
+  sock *s = c->sock;
+  if (!s)
+    return;
 
   if (s->tbuf == NULL)
     cli_write(c);
@@ -485,7 +493,8 @@ cli_tx(sock *s)
 int
 cli_get_command(cli *c)
 {
-  sock *s = c->priv;
+  sock *s = c->sock;
+  ASSERT_DIE(c->sock);
   byte *t = s->rbuf;
   byte *tend = s->rpos;
   byte *d = c->rx_pos;
@@ -524,16 +533,19 @@ cli_rx(sock *s, uint size UNUSED)
   return 0;
 }
 
+#define GLOBAL_CLI_DEBUG (atomic_load_explicit(&global_runtime, memory_order_relaxed)->cli_debug)
+
 static void
 cli_err(sock *s, int err)
 {
-  if (config->cli_debug)
+  if (GLOBAL_CLI_DEBUG)
     {
       if (err)
 	log(L_INFO "CLI connection dropped: %s", strerror(err));
       else
 	log(L_INFO "CLI connection closed");
     }
+
   cli_free(s->data);
 }
 
@@ -541,7 +553,7 @@ static void
 cli_connect_err(sock *s UNUSED, int err)
 {
   ASSERT_DIE(err);
-  if (config->cli_debug)
+  if (GLOBAL_CLI_DEBUG)
     log(L_INFO "Failed to accept CLI connection: %s", strerror(err));
 }
 
@@ -550,7 +562,7 @@ cli_connect(sock *s, uint size UNUSED)
 {
   cli *c;
 
-  if (config->cli_debug)
+  if (GLOBAL_CLI_DEBUG)
     log(L_INFO "CLI connect");
   s->rx_hook = cli_rx;
   s->tx_hook = cli_tx;
@@ -579,7 +591,7 @@ cli_listen(struct cli_config *cf)
   /* Return value intentionally ignored */
   unlink(cf->name);
 
-  if (sk_open_unix(s, cf->name) < 0)
+  if (sk_open_unix(s, &main_birdloop, cf->name) < 0)
   {
     log(L_ERR "Cannot create control socket %s: %m", cf->name);
     goto err;
@@ -1018,20 +1030,25 @@ main(int argc, char **argv)
     dmalloc_debug(0x2f03d00);
 #endif
 
-  parse_args(argc, argv);
-  log_switch(1, NULL, NULL);
-
-  random_init();
+  /* Prepare necessary infrastructure */
+  the_bird_lock();
+  times_update();
   resource_init();
-  timer_init();
+  random_init();
+
+  birdloop_init();
   olock_init();
-  io_init();
   rt_init();
+  io_init();
   if_init();
   mpls_init();
 //  roa_init();
   config_init();
 
+  /* Arguments and logs */
+  parse_args(argc, argv);
+  log_switch(1, NULL, NULL);
+
   uid_t use_uid = get_uid(use_user);
   gid_t use_gid = get_gid(use_group);
 
@@ -1059,6 +1076,8 @@ main(int argc, char **argv)
   if (parse_and_exit)
     exit(0);
 
+  flush_local_pages();
+
   if (!run_in_foreground)
     {
       pid_t pid = fork();
@@ -1074,13 +1093,11 @@ main(int argc, char **argv)
       dup2(0, 2);
     }
 
-  main_thread_init();
-
   write_pid_file();
 
   signal_init();
 
-  config_commit(conf, RECONFIG_HARD, 0);
+  CONFIG_COMMIT(conf, RECONFIG_HARD, 0);
 
   graceful_restart_init();
 
